/*
*/
#include "opcodes.h"
#include "system.h"

#define DEBUG	1
//#define DELAY	0x7FFFF

//=== STRUCTURES ===
#define FLAG_DEFAULT	0x2
#define FLAG_CF	0x001
#define FLAG_PF	0x004
#define FLAG_AF	0x010
#define FLAG_ZF	0x040
#define FLAG_SF	0x080
#define FLAG_TF	0x100
#define FLAG_IF	0x200
#define FLAG_DF	0x400
#define FLAG_OF	0x800

#define STACK(x)	(*(Uint16*)(r->ss*16+(x)))
#if DEBUG
#define DEBUG_S(str) puts(str)
#define DEBUG_H(num) putHex(num)
#define DEBUG_C(ch) putch(ch)
#define DEBUG_F(f,...) printf((f),__VA_ARGS__)
#else
#define DEBUG_S(str)
#define DEBUG_H(num)
#define DEBUG_C(ch)
#define DEBUG_F(f,...)
#endif

extern volatile int kb_press;

//=== PROTOTYPES ===
void RM_Int(int code, int *ax, int *bx, int *cx, int *dx);
void RM_Call(t_regsRM *r);
void RM_DumpRegs(t_regsRM *r);
int genToFromB(t_regsRM *r, Uint8 *data, Uint8 **to, Uint8 **from);
int genToFromW(t_regsRM *r, Uint8 *data, Uint16 **to, Uint16 **from);
Uint16 *seg(t_regsRM *r, int code);
void doCJMP(t_regsRM *r, Uint8 type, Uint16 offset);

//=== CODE ===
/*
Run Realmode interrupt
*/
void RM_Int(int code, int *ax, int *bx, int *cx, int *dx)
{
	struct {
		Uint16	offset;
		Uint16	segment;
	}	*ivt = 0; //*/
	t_regsRM	regs;
	//char	stack[512];
	
	printf("RM_Int: Calling Int 0x%x\n", code);
	
	if(code < 0 || code > 0xFF)
		return;
	
	regs.ax = *ax;	regs.bx = *bx;
	regs.cx = *cx;	regs.dx = *dx;
	
	regs.sp = 512;	regs.ss = 0;
	regs.ds = 0;	regs.es = 0;
	regs.bp = 0;	regs.si = 0;
	regs.di = 0;	regs.flags = FLAG_DEFAULT;
	
	regs.cs = ivt[code].segment;
	regs.ip = ivt[code].offset;
	//regs.cs = 0xF000;
	//regs.ip = 0xFFF0;
	
	//RM_DumpRegs(&regs);
	RM_Call(&regs);
	
	*ax = regs.ax;	*bx = regs.bx;
	*cx = regs.cx;	*dx = regs.dx;
}

/*
Dump Realmode Registers
*/
void RM_DumpRegs(t_regsRM *r)
{
	DEBUG_C('\n');
	DEBUG_F("AX = 0x%x\tBX = 0x%x\tCX = 0x%x\tDX = 0x%x\n", r->ax, r->bx, r->cx, r->dx);
	DEBUG_F("SS = 0x%x\tCS = 0x%x\tDS = 0x%x\tES = 0x%x\n", r->ss, r->cs, r->ds, r->es);
	DEBUG_F("SP = 0x%x\tBP = 0x%x\tSI = 0x%x\tDI = 0x%x\n", r->sp, r->bp, r->si, r->di);
	DEBUG_F("IP = 0x%x\tFLAGS = 0x%x\n", r->ip, r->flags);
}

/*
Call A Realmode function
*/
void RM_Call(t_regsRM *r)
{
	Uint8 *code = (Uint8 *)(r->cs<<4);
	int	pt2;
	Uint16	*toW, *fromW;
	Uint8	*toB, *fromB;
	Uint8	repType = 0;
	Uint16	repStart = 0;
	
	for(;;)
	{
	
		switch(repType) {
		case 0:
			RM_DumpRegs(r);
			break;
		case REP:
			if(r->flags & FLAG_ZF && r->cx--)	r->ip = repStart;
			else	repType = 0;
			break;
		case REPNZ:
			if(r->flags ^ FLAG_ZF && r->cx--)	r->ip = repStart;
			else		repType = 0;
			break;
		case LOOP:
			if(r->cx != 0)	r->ip = repStart;
			else	repType = 0;
			break;
		case LOOPZ:
			if(r->flags & FLAG_ZF)	r->ip = repStart;
			else	repType = 0;
			break;
		case LOOPNZ:
			if(r->flags ^ FLAG_ZF)	r->ip = repStart;
			else	repType = 0;
			break;
		}
		DEBUG_F("[0x%x] %x:%x ", r->cs*16+r->ip, r->cs, r->ip);
		DEBUG_F("(0x%x) ", code[r->ip]);
		//DEBUG_C('0');DEBUG_C('x');DEBUG_H(code[r->ip]>>4); DEBUG_H(code[r->ip]&0xF);DEBUG_C(' ');
		
		switch(code[r->ip++])
		{
		
		//====  ADD ====
		case ADD_RR:	//case ADD_RM:	//Same
			DEBUG_S("ADD (RR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &toB, &fromB);
			if(*toB + *fromB > 0xFF)	r->flags |= FLAG_CF;
			*toB += *fromB;
			break;
		case ADD_RRX:	//case ADD_RMX:	//Same
			DEBUG_S("ADD (RRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, &fromW);
			if(*toW + *fromW > 0xFFFF)	r->flags |= FLAG_CF;
			*toW += *fromW;
			break;
		case ADD_MR:
			DEBUG_S("ADD (MR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &fromB, &toB);
			if(*toB + *fromB > 0xFF)	r->flags |= FLAG_CF;
			*toB += *fromB;
			break;
		case ADD_MRX:
			DEBUG_S("ADD (MRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &fromW, &toW);
			if(*toW + *fromW > 0xFFFF)	r->flags |= FLAG_CF;
			*toW += *fromW;
			break;
		case ADD_AI:
			DEBUG_S("ADD (AI)");
			pt2 = (r->ax&0xFF) + code[r->ip];
			if(pt2 > 0xFF)	r->flags |= FLAG_CF;
			r->ax &= 0xFF00;
			r->ax |= pt2;
			break;
		case ADD_AIX:
			DEBUG_S("ADD (AIX)");
			pt2 = r->ax + *(Uint16*)(code+r->ip);
			if(pt2 > 0xFF)	r->flags |= FLAG_CF;
			r->ax = pt2;
			break;
		
		//====  SUB ====
		case SUB_RR:	//case SUB_RM:	//Same
			DEBUG_S("SUB (RR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &toB, &fromB);
			if(*toB - *fromB < 0)	r->flags |= FLAG_CF;
			*toB -= *fromB;
			break;
		case SUB_RRX:	//case SUB_RMX:	//Same
			DEBUG_S("SUB (RRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, &fromW);
			if(*toW - *fromW < 0)	r->flags |= FLAG_CF;
			*toW -= *fromW;
			break;
		case SUB_MR:
			DEBUG_S("SUB (MR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &fromB, &toB);
			if(*toB + *fromB < 0)	r->flags |= FLAG_CF;
			*toB -= *fromB;
			break;
		case SUB_MRX:
			DEBUG_S("SUB (MRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &fromW, &toW);
			if(*toW + *fromW < 0)	r->flags |= FLAG_CF;
			*toW -= *fromW;
			break;
		case SUB_AI:
			DEBUG_S("SUB (AI)");
			pt2 = (r->ax&0xFF) - code[r->ip];
			if(pt2 < 0)	r->flags |= FLAG_CF;
			r->ax &= 0xFF00;
			r->ax |= pt2;
			break;
		case SUB_AIX:
			DEBUG_S("SUB (AIX)");
			pt2 = r->ax - *(Uint16*)(code+r->ip);
			if(pt2 < 0)	r->flags |= FLAG_CF;
			r->ax = pt2;
			break;
		
		//AND Family
		case AND_RR:	DEBUG_S("AND (RR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &toB, &fromB);
			*toB &= *fromB;
			break;
		case AND_RRX:	DEBUG_S("AND (RRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, &fromW);
			*toW &= *fromW;
			break;
		case AND_MR:	DEBUG_S("AND (MR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &fromB, &toB);
			*toB &= *fromB;
			break;
		case AND_MRX:	DEBUG_S("AND (MRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &fromW, &toW);
			*toW &= *fromW;
			break;
		case AND_AI:	DEBUG_S("AND (AI) AL 0x"); DEBUG_H(code[r->ip]);
			r->ax &= 0xFF00|code[r->ip];
			r->ip++;
			break;
		case AND_AIX:	DEBUG_S("AND (AIX) AX "); DEBUG_H( *(Uint16*)(code+r->ip) );
			r->ax = *(Uint16*)(code+r->ip);
			r->ip++;
			break;
		
		//OR Family
		case OR_RR:		DEBUG_S("OR (RR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &toB, &fromB);
			*toB |= *fromB;
			break;
		case OR_RRX:	DEBUG_S("OR (RRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, &fromW);
			*toW |= *fromW;
			break;
		case OR_MR:		DEBUG_S("OR (MR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &fromB, &toB);
			*toB |= *fromB;
			break;
		case OR_MRX:	DEBUG_S("OR (MRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &fromW, &toW);
			*toW |= *fromW;
			break;
		case OR_AI:		DEBUG_S("OR (AI) AL 0x"); DEBUG_H(code[r->ip]);
			r->ax |= code[r->ip]&0xFF;
			r->ip++;
			break;
		case OR_AIX:	DEBUG_S("OR (AIX) AX "); DEBUG_H( *(Uint16*)(code+r->ip) );
			r->ax |= *(Uint16*)(code+r->ip);
			r->ip++;
			break;
		
		
		//==== COMPOUND 1st Bytes ====
		case 0x0F:
			pt2 = code[r->ip];
			switch(pt2) {
			
			//--- PUSH Segment --- (100sss000)
			case 0x80:	/*10 000 001 */	case 0x88:	/*10 001 001 */
			case 0x90:	/*10 010 001 */	case 0x98:	/*10 011 001 */
				fromW = seg(r, (code[r->ip]&0x38)>>3);
				STACK(r->sp-=2) = *fromW;
				break;
			
			//--- POP Segment --- (100sss001)
			case 0x81:	/*10 000 001 */	case 0x89:	/*10 001 001 */
			case 0x91:	/*10 010 001 */	case 0x99:	/*10 011 001 */
				toW = seg(r, (code[r->ip]&0x38)>>3);
				*toW = STACK(r->sp);	r->sp += 2;
				break;
				
			//--- Near Jump --- (1000cccc)
			//TODO: Check why collision at 80,81,88,89 exists
			//HACK: Ingnoring Overflow and No Overflow
			/*case 0x80:	case 0x81:*/	case 0x82:	case 0x83:
			case 0x84:	case 0x85:	case 0x86:	case 0x87:
			/*case 0x88:	case 0x89:*/	case 0x8A:	case 0x8B:
			case 0x8C:	case 0x8D:	case 0x8E:	case 0x8F:
				r->ip++;
				doCJMP(r, pt2&0xF, *(Uint16*)(code+r->ip));
				r->ip+=2;
				break;
			
			}
			break;
			
		case 0x80:
			pt2 = (code[r->ip]>>3)&7;	//Get Flag (rrr)
			switch(pt2){
			case 0:	DEBUG_S("ADD (RI)");	break;
			case 4:	DEBUG_S("AND (RI)");	break;
			case 7: DEBUG_S("CMP (RI)");	break;
			default:	putHex(pt2);	break;
			}
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), NULL, &toB);
			//DEBUG_F("*toB = 0x%x\n", *toB);
			DEBUG_S(" 0x");DEBUG_H(code[r->ip]);
			switch(pt2) {
			case 0:
				*toB += code[r->ip];
				break;
			case 4:
				*toB &= code[r->ip];
				break;
			case 7:
				if(*toB == code[r->ip])	r->flags |= FLAG_ZF;
				else	r->flags &= ~FLAG_ZF;
				if(*toB + code[r->ip] > 0xFF)	r->flags |= FLAG_CF;	//Set Carry Bit
				else	r->flags &= ~FLAG_CF;
				break;
			}
			if(*toB < code[r->ip])	r->flags |= FLAG_CF;	//Set Carry Bit
			else	r->flags &= ~FLAG_CF;
			r->ip += 1;
			break;
		case 0x81:	//Also CMP_RIX
			pt2 = (code[r->ip]&0x38)>>3;	//Get Flag (rrr)
			switch(pt2){
			case 0:	DEBUG_S("ADD (RIX)");	break;
			case 4:	DEBUG_S("AND (RIX)");	break;
			case 7: DEBUG_S("CMP (RIX)");	break;
			default:	putBin(pt2);	break;
			}
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);	//Get Register Value
			DEBUG_S(" 0x");DEBUG_H(*(Uint16 *)(code+r->ip));	//Print Immediate Valiue
			switch(pt2) {
			case 0:
				*toW += *(Uint16 *)(code+r->ip);
				break;
			case 4:
				*toW &= *(Uint16 *)(code+r->ip);
				break;
			case 7:
				if(*toW == *(Uint16 *)(code+r->ip))	r->flags |= FLAG_ZF;
				else	r->flags &= ~FLAG_ZF;
				if(*toW + *(Uint16 *)(code+r->ip) > 0xFFFF)	r->flags |= FLAG_CF;	//Set Carry Bit
				else	r->flags &= ~FLAG_CF;
				break;
			}
			if(*toW < *(Uint16 *)(code+r->ip))	r->flags |= FLAG_CF;	//Set Carry Bit
			else	r->flags &= ~FLAG_CF;
			r->ip += 2;
			break;
		
		// Immediate Maths Functions
		case 0x82:	//ADD_RI8
			pt2 = (code[r->ip]&0x38)>>3;	//Get Flag (rrr)
			switch(pt2) {
			case 0:	DEBUG_S("ADD (RI8)");	break;
			case 5:	DEBUG_S("SUB (RI8)");	break;
			case 7:	DEBUG_S("CMP (RI8)");	break;
			default:	DEBUG_S("UNIMPLEMENTED 0x82");	break;
			}
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), NULL, &toB);	//Get Register Value
			DEBUG_S(" 0x");DEBUG_H(code[r->ip]);
			switch(pt2) {
			case 0:
				*toB += code[r->ip];
				break;
			case 5:
				*toB -= code[r->ip];
				break;
			case 7:
				if(*toB == code[r->ip])	r->flags |= FLAG_ZF;
				break;
			}
			if(*toB + code[r->ip] > 0xFF)	r->flags |= FLAG_CF;	//Set Carry Bit
			r->ip += 1;
			break;
		case 0x83:	//ADD_RI8X
			pt2 = (code[r->ip]&0x38)>>3;	//Get Flag (rrr)
			switch(pt2) {
			case 0:	DEBUG_S("ADD (RI8X)");	break;
			case 5:	DEBUG_S("SUB (RI8X)");	break;
			case 7:	DEBUG_S("CMP (RI8X)");	break;
			default:	DEBUG_S("UNIMPLEMENTED 0x83");	break;
			}
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);	//Get Register Value
			DEBUG_S(" 0x");DEBUG_H(code[r->ip]);
			switch(pt2) {
			case 0:
				*toW += (code[r->ip]&0x80?0xFF00:0)|code[r->ip];
				break;
			case 5:
				*toW -= (code[r->ip]&0x80?0xFF00:0)|code[r->ip];
				break;
			case 7:
				if( *toW == ((code[r->ip]&0x80?0xFF00:0)|code[r->ip]) )
					r->flags |= FLAG_ZF;
				break;
			}
			if(*toW + code[r->ip] > 0xFF)	r->flags |= FLAG_CF;	//Set Carry Bit
			r->ip += 1;
			break;
		
		case 0xC0:	//Register I8
			switch((code[r->ip]&0x38)>>3) {
			case 4:
				DEBUG_S("SHL (RI8)");
				r->ip += genToFromB(r, (Uint8*)(code+r->ip), NULL, &toB);
				*toB <<= code[r->ip];
				r->ip++;
				break;
			case 5:
				DEBUG_S("SHR (RI8)");
				r->ip += genToFromB(r, (Uint8*)(code+r->ip), NULL, &toB);
				*toB >>= code[r->ip];
				r->ip++;
				break;
			}
			break;
		case 0xC1:	//Register I8 Extended
			switch((code[r->ip]&0x38)>>3) {
			case 4:
				DEBUG_S("SHL (RI8X)");
				r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);
				*toW <<= code[r->ip];
				r->ip++;
				break;
			case 5:
				DEBUG_S("SHR (RI8X)");
				r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);
				*toW >>= code[r->ip];
				r->ip++;
				break;
			}
			break;
		
		case 0xD0:	//Register
			switch((code[r->ip]&0x38)>>3) {
			case 4:
				DEBUG_S("SHL (R1)");
				r->ip += genToFromB(r, (Uint8*)(code+r->ip), NULL, &toB);
				*toB <<= 1;
				break;
			case 5:
				DEBUG_S("SHR (R1)");
				r->ip += genToFromB(r, (Uint8*)(code+r->ip), NULL, &toB);
				*toB >>= 1;
				break;
			}
			break;
		case 0xD1:	//Register Extended
			switch((code[r->ip]&0x38)>>3) {
			case 4:
				DEBUG_S("SHL (R1X)");
				r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);
				*toW <<= 1;
				break;
			case 5:
				DEBUG_S("SHR (R1X)");
				r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);
				*toW >>= 1;
				break;
			}
			break;
		
		case 0xD2:	//Register CL
			switch((code[r->ip]&0x38)>>3) {
			case 4:
				DEBUG_S("SHL (RCl)");
				r->ip += genToFromB(r, (Uint8*)(code+r->ip), NULL, &toB);
				*toB <<= 1;
				break;
			case 5:
				DEBUG_S("SHR (RCl)");
				r->ip += genToFromB(r, (Uint8*)(code+r->ip), NULL, &toB);
				*toB >>= 1;
				break;
			}
			break;
		case 0xD3:	//Register CL Extended
			switch((code[r->ip]&0x38)>>3) {
			case 4:
				DEBUG_S("SHL (RClX)");
				r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);
				*toW <<= r->cx&0xFF;
				break;
			case 5:
				DEBUG_S("SHR (RClX)");
				r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);
				*toW >>= r->cx&0xFF;
				break;
			}
			break;
			
		case 0xF6:	//Register Immidate
			switch((code[r->ip]&0x38)>>3) {
			case 0:
				DEBUG_S("TEST (RI)");
				r->ip += genToFromB(r, (Uint8*)(code+r->ip), NULL, &toB);
				DEBUG_S(" 0x"); DEBUG_H(code[r->ip]);
				pt2 = *toB & code[r->ip];
				r->flags &= ~(FLAG_CF|FLAG_OF|FLAG_ZF|FLAG_SF|FLAG_PF);	//Unset Clear,Overflow,Zero,Sign,Parity
				if(pt2 == 0)	r->flags |= FLAG_ZF;	//Zero Flag
				if(pt2 & 0x80)	r->flags |= FLAG_SF;	//Sign Flag
				while( pt2 >>= 1 )
					if(pt2&1)	r->flags ^= FLAG_PF;	//Set Parity
				r->ip++;
				break;
			}
			break;
		case 0xF7:	//Register Immidate Extended
			switch((code[r->ip]&0x38)>>3) {
			case 0:
				DEBUG_S("TEST (RIX)");
				r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);
				DEBUG_S(" 0x"); DEBUG_H(*(Uint16*)(code+r->ip));
				pt2 = *toW & *(Uint16*)(code+r->ip);
				r->flags &= ~(FLAG_CF|FLAG_OF|FLAG_ZF|FLAG_SF|FLAG_PF);	//Unset Clear,Overflow,Zero,Sign,Parity
				if(pt2 == 0)	r->flags |= FLAG_ZF;	//Zero Flag
				if(pt2 & 0x8000)	r->flags |= FLAG_SF;	//Sign Flag
				while( pt2 >>= 1 )
					if(pt2&1)	r->flags ^= FLAG_PF;	//Set Parity
				r->ip++;
				break;
			}
			break;
			
		case 0xFE:	//Register
			pt2 = (code[r->ip]&0x38)>>3;
			switch(pt2) {
			case 0:
				DEBUG_S("INC (R)");
				r->ip += genToFromB(r, (Uint8*)(code+r->ip), NULL, &toB);	//Get Register Value
				*toB ++;
				break;
			case 1:
				DEBUG_S("DEC (R)");
				r->ip += genToFromB(r, (Uint8*)(code+r->ip), NULL, &toB);	//Get Register Value
				*toB --;
				break;
			}
			break;
		case 0xFF:	//Register Extended
			pt2 = (code[r->ip]&0x38)>>3;
			switch(pt2) {
			case 0:
				DEBUG_S("INC (RX)");
				r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);	//Get Register Value
				*toW ++;
				break;
			case 1:
				DEBUG_S("DEC (RX)");
				r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);	//Get Register Value
				*toW --;
				break;
			case 2:
				DEBUG_S("CALL (RX) NEAR");
				r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);	//Get Register Value
				STACK(r->sp-=2) = r->ip;
				r->ip += *toW;
				break;
			case 3:
				DEBUG_S("CALL (MX) FAR --NI--");
				break;
			case 4:
				DEBUG_S("JMP (RX) NEAR --NI--");
				break;
			case 5:
				DEBUG_S("JMP (MX) FAR --NI--");
				break;
			case 6:
				DEBUG_S("PUSH (RX)");
				r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);	//Get Register Value
				STACK(r->sp-=2) = *toW;
				break;
			}
			break;
			
		//TEST Family
		case TEST_RR:	DEBUG_S("TEST (RR)");	//Test Register
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &toB, &fromB);
			pt2 = *toB & *fromB;
			r->flags &= ~(FLAG_CF|FLAG_OF|FLAG_ZF|FLAG_SF|FLAG_PF);
			if(pt2 == 0)	r->flags |= FLAG_ZF;
			if(pt2 & 0x80)	r->flags |= FLAG_SF;
			while( pt2 >>= 1 )
				if(pt2&1)	r->flags ^= FLAG_PF;	//Set Parity
			break;
		case TEST_RRX:	DEBUG_S("TEST (RRX)");	//Test Register Extended
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, &fromW);
			pt2 = *toW & *fromW;
			r->flags &= ~(FLAG_CF|FLAG_OF|FLAG_ZF|FLAG_SF|FLAG_PF);
			if(pt2 == 0)	r->flags |= FLAG_ZF;
			if(pt2 & 0x8000)	r->flags |= FLAG_SF;
			while( pt2 >>= 1 )
				if(pt2&1)	r->flags ^= FLAG_PF;	//Set Parity
			break;
			
		//XOR Family
		case XOR_RR:
			DEBUG_S("XOR (RR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &toB, &fromB);
			*toB ^= *fromB;
			break;
		case XOR_RRX:
			DEBUG_S("XOR (RRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, &fromW);
			*toW ^= *fromW;
			break;
		case XOR_MR:
			DEBUG_S("XOR (MR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &toB, &fromB);
			*toB ^= *fromB;
			break;
		case XOR_MRX:
			DEBUG_S("XOR (MRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, &fromW);
			*toW ^= *fromW;
			break;
		case XOR_AI:
			DEBUG_S("XOR (AI)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &toB, NULL);
			*toB ^= code[r->ip];
			r->ip++;
			break;
		case XOR_AIX:
			DEBUG_S("XOR (AIX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, NULL);
			*toB ^= *(Uint16*)(code+r->ip);
			r->ip += 2;
			break;
		
		//CLI - Ignored
		case CLI:
			DEBUG_S("CLI");
			break;
			
		//CMP - Unknown Implementation
		// TODO: Check and Implement
		case CMP_RR:
			DEBUG_S("CMP (RR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &toB, &fromB);
			if(*toB + *fromB > 0xFF)	r->flags |= FLAG_CF;
			if(*toB == *fromB)	r->flags |= FLAG_ZF;
			break;
		case CMP_RRX:
			DEBUG_S("CMP (RRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, &fromW);
			if(*toW + *fromW > 0xFF)	r->flags |= FLAG_CF;
			if(*toW == *fromW)	r->flags |= FLAG_ZF;
			break;
		case CMP_MR:
			DEBUG_S("CMP (MR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &fromB, &toB);
			if(*toB + *fromB > 0xFF)	r->flags |= FLAG_CF;
			if(*toB == *fromB)	r->flags |= FLAG_ZF;
			break;
		case CMP_MRX:
			DEBUG_S("CMP (MRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &fromW, &toW);
			if(*toW + *fromW > 0xFF)	r->flags |= FLAG_CF;
			if(*toW == *fromW)	r->flags |= FLAG_ZF;
			break;
		case CMP_AI:
			DEBUG_S("CMP (AI) AL 0x");	DEBUG_H(code[r->ip]);
			if((r->ax&0xFF) + code[r->ip] > 0xFF)	r->flags |= FLAG_CF;
			if((r->ax&0xFF) == code[r->ip])	r->flags |= FLAG_ZF;
			r->ip ++;
			break;
		case CMP_AIX:
			DEBUG_S("CMP (AIX) AX 0x");DEBUG_H(*(Uint16*)(code+r->ip));
			if(r->ax + *(Uint16*)(code+r->ip) > 0xFFFF)	r->flags |= FLAG_CF;
			if(r->ax == *(Uint16*)(code+r->ip))	r->flags |= FLAG_ZF;
			r->ip += 2;
			break;
		
		//DEC Family
		case DEC_A:		DEBUG_S("DEC AX");	r->ax --;	break;
		case DEC_B:		DEBUG_S("DEC BX");	r->bx --;	break;
		case DEC_C:		DEBUG_S("DEC CX");	r->cx --;	break;
		case DEC_D:		DEBUG_S("DEC DX");	r->dx --;	break;
		case DEC_Sp:	DEBUG_S("DEC SP");	r->sp --;	break;
		case DEC_Bp:	DEBUG_S("DEC BP");	r->bp --;	break;
		case DEC_Si:	DEBUG_S("DEC SI");	r->si --;	break;
		case DEC_Di:	DEBUG_S("DEC DI");	r->di --;	break;
		//INC Family
		case INC_A:		DEBUG_S("INC AX");	r->ax ++;	break;
		case INC_B:		DEBUG_S("INC BX");	r->bx ++;	break;
		case INC_C:		DEBUG_S("INC CX");	r->cx ++;	break;
		case INC_D:		DEBUG_S("INC DX");	r->dx ++;	break;
		case INC_Sp:	DEBUG_S("INC SP");	r->sp ++;	break;
		case INC_Bp:	DEBUG_S("INC BP");	r->bp ++;	break;
		case INC_Si:	DEBUG_S("INC SI");	r->si ++;	break;
		case INC_Di:	DEBUG_S("INC DI");	r->di ++;	break;
		
		//IN Family
		case IN_AI:
			DEBUG_S("IN (AI) 0x"); DEBUG_H(*(Uint8*)(code+r->ip)); DEBUG_S(" AL");
			r->ax &= 0xFF00;
			r->ax |= inportb( *(Uint8*)(code+r->ip) );
			r->ip += 1;
			break;
		case IN_AIX:
			DEBUG_S("IN (AIX) 0x"); DEBUG_H(*(Uint16*)(code+r->ip)); DEBUG_S(" AX");
			r->ax = inportw( *(Uint16*)(code+r->ip) );
			r->ip += 2;
			break;
		case IN_ADx:
			DEBUG_S("IN (ADx) DX AL");
			r->ax &= 0xFF00;
			r->ax |= inportb(r->dx);
			break;
		case IN_ADxX:
			DEBUG_S("IN (ADxX) DX AX");
			r->ax = inportw(r->dx);
			break;
		//OUT Family
		case OUT_IA:
			DEBUG_S("OUT (IA) 0x"); DEBUG_H(*(Uint8*)(code+r->ip)); DEBUG_S(" AL");
			outportb( *(Uint8*)(code+r->ip), r->ax&0xFF );
			r->ip += 1;
			break;
		case OUT_IAX:
			DEBUG_S("OUT (IAX) 0x"); DEBUG_H(*(Uint16*)(code+r->ip)); DEBUG_S(" AX");
			outportw( *(Uint16*)(code+r->ip), r->ax );
			r->ip += 2;
			break;
		case OUT_DxA:
			DEBUG_S("OUT (DxA) DX AL");
			outportb( r->dx, r->ax&0xFF );
			break;
		case OUT_DxAX:
			DEBUG_S("OUT (DxAX) DX AX");
			outportb( r->dx, r->ax );
			break;
		
		//INT Family
		case INT3:
			DEBUG_S("INT 3");
			//debugRealMode(r);
			
			break;
		case INT_I:
			DEBUG_S("INT 0x");	DEBUG_H(code[r->ip]);
			r->ip++;
			break;
		case IRET:
			DEBUG_S("IRET\n");
			return;
			break;
			
		//MOV Family
		case MOV_MoA:	DEBUG_S("MOV (MoA)");	//Memory Offset from AL
			pt2 = *(Uint16*)(code+r->ip);
			DEBUG_S(" DS:0x");	DEBUG_H(pt2);	DEBUG_S(" AL");
			*(Uint8*)(r->ds*16+pt2) = r->ax&0xFF;
			r->ip += 2;
			break;
		case MOV_MoAX:	//Memory Offset from AX
			DEBUG_S("MOV (MoAX)");
			pt2 = *(Uint16*)(code+r->ip);
			DEBUG_S(" DS:0x");	DEBUG_H(pt2);	DEBUG_S(" AX");
			*(Uint16*)(r->ds*16+pt2) = r->ax;
			r->ip += 2;
		case MOV_AMo:	//Memory Offset to AL
			DEBUG_S("MOV (AMo)");
			pt2 = *(Uint16*)(code+r->ip);
			DEBUG_S("AX DS:0x");	DEBUG_H(pt2);
			r->ax &= 0xFF00;
			r->ax |= *(Uint8*)(r->ds*16+pt2);
			r->ip += 2;
			break;
		case MOV_AMoX:	//Memory Offset to AX
			DEBUG_S("MOV (AMoX)");
			pt2 = *(Uint16*)(code+r->ip);
			DEBUG_S(" AX DS:0x");	DEBUG_H(pt2);
			r->ax = *(Uint16*)(r->ds*16+pt2);
			r->ip += 2;
			break;
		case MOV_MI:
			DEBUG_S("MOV (MI)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &toB, NULL);
			*toB = code[r->ip];
			r->ip ++;
			break;
		case MOV_MIX:
			DEBUG_S("MOV (MIX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, NULL);
			*toW = *(Uint16*)(code+r->ip);
			r->ip += 2;
			break;
		case MOV_RR:
			DEBUG_S("MOV (RR)");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &toB, &fromB);
			*toB = *fromB;
			break;
		case MOV_RRX:
			DEBUG_S("MOV (RRX)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, &fromW);
			*toW = *fromW;
			break;
		case MOV_MR:	DEBUG_S("MOV (MR)REV");
			r->ip += genToFromB(r, (Uint8*)(code+r->ip), &fromB, &toB);
			*toB = *fromB;
			break;
		case MOV_MRX:	DEBUG_S("MOV (MRX)REV");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &fromW, &toW);
			*toW = *fromW;
			break;
		case MOV_RI_AL:	DEBUG_S("MOV (RI) AL 0x");	DEBUG_H(code[r->ip]);
			r->ax &= 0xFF00;	r->ax |= code[r->ip];
			r->ip ++;
			break;
		case MOV_RI_BL:	DEBUG_S("MOV (RI) BL 0x");	DEBUG_H(code[r->ip]);
			r->bx &= 0xFF00;	r->bx |= code[r->ip];
			r->ip ++;
			break;
		case MOV_RI_CL:	DEBUG_S("MOV (RI) CL 0x");	DEBUG_H(code[r->ip]);
			r->cx &= 0xFF00;	r->cx |= code[r->ip];
			r->ip ++;
			break;
		case MOV_RI_DL:	DEBUG_S("MOV (RI) DL 0x");	DEBUG_H(code[r->ip]);
			r->dx &= 0xFF00;	r->dx |= code[r->ip];
			r->ip ++;
			break;
		case MOV_RI_AH:	DEBUG_S("MOV (RI) AH 0x");	DEBUG_H(code[r->ip]);
			r->ax &= 0xFF;	r->ax |= code[r->ip]<<8;
			r->ip ++;
			break;
		case MOV_RI_BH:	DEBUG_S("MOV (RI) BH 0x");	DEBUG_H(code[r->ip]);
			r->bx &= 0xFF;	r->bx |= code[r->ip]<<8;
			r->ip ++;
			break;
		case MOV_RI_CH:	DEBUG_S("MOV (RI) CH 0x");	DEBUG_H(code[r->ip]);
			r->cx &= 0xFF;	r->cx |= code[r->ip]<<8;
			r->ip ++;
			break;
		case MOV_RI_DH:	DEBUG_S("MOV (RI) DH 0x");	DEBUG_H(code[r->ip]);
			r->dx &= 0xFF;	r->dx |= code[r->ip]<<8;
			r->ip ++;
			break;
		case MOV_RI_AX:	DEBUG_S("MOV (RI) AX 0x");	DEBUG_H(*(Uint16*)(code+r->ip));
			r->ax = *(Uint16*)(code+r->ip);		r->ip += 2;
			break;
		case MOV_RI_BX:	DEBUG_S("MOV (RI) BX 0x");	DEBUG_H(*(Uint16*)(code+r->ip));
			r->bx = *(Uint16*)(code+r->ip);		r->ip += 2;
			break;
		case MOV_RI_CX:	DEBUG_S("MOV (RI) CX 0x");	DEBUG_H(*(Uint16*)(code+r->ip));
			r->cx = *(Uint16*)(code+r->ip);		r->ip += 2;
			break;
		case MOV_RI_DX:	DEBUG_S("MOV (RI) DX 0x");	DEBUG_H(*(Uint16*)(code+r->ip));
			r->dx = *(Uint16*)(code+r->ip);		r->ip += 2;
			break;
		case MOV_RI_SP:	DEBUG_S("MOV (RI) SP 0x");	DEBUG_H(*(Uint16*)(code+r->ip));
			r->sp = *(Uint16*)(code+r->ip);		r->ip += 2;
			break;
		case MOV_RI_BP:	DEBUG_S("MOV (RI) BP 0x");	DEBUG_H(*(Uint16*)(code+r->ip));
			r->bp = *(Uint16*)(code+r->ip);		r->ip += 2;
			break;
		case MOV_RI_SI:	DEBUG_S("MOV (RI) SI 0x");	DEBUG_H(*(Uint16*)(code+r->ip));
			r->si = *(Uint16*)(code+r->ip);		r->ip += 2;
			break;
		case MOV_RI_DI:	DEBUG_S("MOV (RI) DI 0x");	DEBUG_H(*(Uint16*)(code+r->ip));
			r->di = *(Uint16*)(code+r->ip);		r->ip += 2;
			break;
		
		case MOV_RS:
			DEBUG_S("MOV (RS)");
			fromW = seg(r, (code[r->ip]>>3)&7);
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);
			*toW = *fromW;
			break;
			
		case MOV_SR:
			DEBUG_S("MOV (SR)");
			toW = seg(r, (code[r->ip]>>3)&7);
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &fromW);
			*toW = *fromW;
			break;
			
			
		//JMP Family
		case JMP_S:	//Short Jump
			DEBUG_S("JMP (S) .+0x");DEBUG_H(code[r->ip]);
			r->ip += *(Uint8*)(code+r->ip) + 1;
			break;
		case JMP_N:	//Near Jump
			DEBUG_S("JMP (S) .+0x");DEBUG_H( *(Uint16*)(code+r->ip) );
			r->ip += *(Uint16*)(code+r->ip) + 2;
			break;
		case JMP_F:	//Near Jump
			DEBUG_S("JMP FAR ");DEBUG_H( *(Uint16*)(code+r->ip+2) );
			DEBUG_C(':');DEBUG_H( *(Uint16*)(code+r->ip) );
			r->cs = *(Uint16*)(code+r->ip+2);
			r->ip = *(Uint16*)(code+r->ip);
			code = (Uint8*)(r->cs<<4);
			break;
		
		//XCHG Family
		case XCHG_AA:	//NOP 0x90
			DEBUG_S("NOP");
			break;
		case XCHG_AB:
			DEBUG_S("XCHG AX BX");
			pt2 = r->ax;	r->ax = r->bx;	r->bx = pt2;
			break;
		case XCHG_AC:
			DEBUG_S("XCHG AX CX");
			pt2 = r->ax;	r->ax = r->cx;	r->cx = pt2;
			break;
		case XCHG_AD:
			DEBUG_S("XCHG AX DX");
			pt2 = r->ax;	r->ax = r->dx;	r->dx = pt2;
			break;
		case XCHG_ASp:
			DEBUG_S("XCHG AX SP");
			pt2 = r->ax;	r->ax = r->sp;	r->sp = pt2;
			break;
		case XCHG_ABp:
			DEBUG_S("XCHG AX BP");
			pt2 = r->ax;	r->ax = r->bp;	r->bp = pt2;
			break;
		case XCHG_ASi:
			DEBUG_S("XCHG AX SI");
			pt2 = r->ax;	r->ax = r->si;	r->si = pt2;
			break;
		case XCHG_ADi:
			DEBUG_S("XCHG AX DI");
			pt2 = r->ax;	r->ax = r->di;	r->di = pt2;
			break;
		case XCHG_RR:
			DEBUG_S("XCHG (RR)");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, &fromW);
			pt2 = *toW;		*toW = *fromW;	*fromW = pt2;
			break;
			
		//PUSH Family
		case PUSHF:
			DEBUG_S("PUSHF");
			STACK(r->sp-=2) = r->flags;
			break;
		case PUSHA:
			DEBUG_S("PUSHA");
			STACK(r->sp-=2) = r->ax;	STACK(r->sp-=2) = r->bx;
			STACK(r->sp-=2) = r->cx;	STACK(r->sp-=2) = r->dx;
			STACK(r->sp-=2) = r->sp;	STACK(r->sp-=2) = r->bp;
			STACK(r->sp-=2) = r->si;	STACK(r->sp-=2) = r->di;
			//STACK(r->sp-=2) = r->ss;	STACK(r->sp-=2) = r->cs;
			//STACK(r->sp-=2) = r->ds;	STACK(r->sp-=2) = r->es;
			break;
		case PUSH_AX:	DEBUG_S("PUSH AX");	STACK(r->sp-=2) = r->ax;	break;
		case PUSH_BX:	DEBUG_S("PUSH BX");	STACK(r->sp-=2) = r->bx;	break;
		case PUSH_CX:	DEBUG_S("PUSH CX");	STACK(r->sp-=2) = r->cx;	break;
		case PUSH_DX:	DEBUG_S("PUSH DX");	STACK(r->sp-=2) = r->dx;	break;
		case PUSH_SP:	DEBUG_S("PUSH Sp");	STACK(r->sp) = r->sp;	r->sp-=2;	break;
		case PUSH_BP:	DEBUG_S("PUSH BP");	STACK(r->sp-=2) = r->bp;	break;
		case PUSH_SI:	DEBUG_S("PUSH SI");	STACK(r->sp-=2) = r->si;	break;
		case PUSH_DI:	DEBUG_S("PUSH DI");	STACK(r->sp-=2) = r->di;	break;
		case PUSH_ES:	DEBUG_S("PUSH ES");	STACK(r->sp-=2) = r->es;	break;
		case PUSH_CS:	DEBUG_S("PUSH CS");	STACK(r->sp-=2) = r->cs;	break;
		case PUSH_SS:	DEBUG_S("PUSH SS");	STACK(r->sp-=2) = r->ss;	break;
		case PUSH_DS:	DEBUG_S("PUSH DS");	STACK(r->sp-=2) = r->ds;	break;
		case PUSH_I8:
			DEBUG_S("PUSH (I8) 0x"); DEBUG_H(code[r->ip]);
			STACK(r->sp-=2) = code[r->ip];
			r->ip++;
			break;
		case PUSH_I:
			DEBUG_S("PUSH (I) "); DEBUG_H( *(Uint16*)(code+r->ip) );
			STACK(r->sp-=2) = *(Uint16*)(code+r->ip);
			r->ip+=2;
			break;		

		//POP Family
		case POPF:
			DEBUG_S("POPF");
			r->flags = STACK(r->sp);	r->sp += 2;
			break;
		case POPA:
			DEBUG_S("POPA");
			//r->es = STACK(r->sp); r->sp+=2;	r->ds = STACK(r->sp); r->sp+=2;
			//r->cs = STACK(r->sp); r->sp+=2;	r->ss = STACK(r->sp); r->sp+=2;
			r->di = STACK(r->sp); r->sp+=2;	r->si = STACK(r->sp); r->sp+=2;
			r->bp = STACK(r->sp); r->sp+=2;	r->sp = STACK(r->sp); r->sp+=2;
			r->dx = STACK(r->sp); r->sp+=2;	r->cx = STACK(r->sp); r->sp+=2;
			r->bx = STACK(r->sp); r->sp+=2;	r->ax = STACK(r->sp); r->sp+=2;
			break;
		case POP_AX:	DEBUG_S("POP AX");	r->ax = STACK(r->sp); r->sp+=2;	break;
		case POP_BX:	DEBUG_S("POP BX");	r->bx = STACK(r->sp); r->sp+=2;	break;
		case POP_CX:	DEBUG_S("POP CX");	r->cx = STACK(r->sp); r->sp+=2;	break;
		case POP_DX:	DEBUG_S("POP DX");	r->dx = STACK(r->sp); r->sp+=2;	break;
		case POP_SP:	DEBUG_S("POP SP");	r->ax = STACK(r->sp+=2);	break;
		case POP_BP:	DEBUG_S("POP BP");	r->bp = STACK(r->sp); r->sp+=2;	break;
		case POP_SI:	DEBUG_S("POP SI");	r->si = STACK(r->sp); r->sp+=2;	break;
		case POP_DI:	DEBUG_S("POP DI");	r->di = STACK(r->sp); r->sp+=2;	break;
		case POP_ES:	DEBUG_S("POP ES");	r->es = STACK(r->sp); r->sp+=2;	break;
//		case POP_CS:	DEBUG_S("POP CS");	r->cs = STACK(r->sp); r->sp+=2;	break;
		case POP_SS:	DEBUG_S("POP SS");
			r->ss = STACK(r->sp);	r->sp+=2;
			break;
		case POP_DS:	DEBUG_S("POP DS");	r->ds = STACK(r->sp); r->sp+=2;	break;
		
		//CALL Family
		/*case CALL_MF:	//MF, MN, R
			DEBUG_S("CALL (MF) ");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), NULL, &toW);
			STACK(r->sp-=2) = r->ip;
			r->ip += *toW;
			break;*/
		case CALL_N:
			DEBUG_S("CALL (N) .+0x"); DEBUG_H(*(Uint16*)(code+r->ip));
			r->ip+=2;
			STACK(r->sp-=2) = r->ip;
			r->ip += *(Uint16*)(code+r->ip-2);
			break;
		case CALL_F:
			DEBUG_S("CALL (F) "); DEBUG_H(*(Uint16*)(code+r->ip+2));
			DEBUG_C(':');	DEBUG_H(*(Uint16*)(code+r->ip));
			r->ip+=4;
			STACK(r->sp-=2) = r->cs;
			STACK(r->sp-=2) = r->ip;
			r->cs = *(Uint16*)(code+r->ip-2);
			r->ip = *(Uint16*)(code+r->ip-4);
			code = (Uint8*)(r->cs<<4);
			break;
		case RET_N:
			DEBUG_S("RET (N)");
			r->ip = STACK(r->sp); r->sp+=2;
			break;
		case RET_F:
			DEBUG_S("RET (F)");
			r->ip = STACK(r->sp); r->sp+=2;
			r->cs = STACK(r->sp); r->sp+=2;
			code = (Uint8*)(r->cs<<4);
			break;
		
		//STOS Functions
		case STOSB:
			*(Uint8*)(r->es*16+r->di) = r->ax & 0xFF;
			r->di ++;
			break;
		case STOSW:
			*(Uint16*)(r->es*16+r->di) = r->ax;
			r->di += 2;
			break;
		
		//Loops/Repeats
		case REP:
		case REPNZ:
		case LOOP:
		case LOOPZ:
		case LOOPNZ:
			repType = code[r->ip-1];
			repStart = r->ip;
			break;
		
		//Misc
		case LEA:
			DEBUG_S("LEA");
			r->ip += genToFromW(r, (Uint8*)(code+r->ip), &toW, &fromW);
			if((Uint32) fromW > 0x100000) {
				DEBUG_S("\nERROR: Memory Address Expected\n");
				return;
			}
			*toW = (Uint32) fromW & 0xFFFF;
			break;
		case CLD:
			DEBUG_S("CLD");
			r->flags &= ~FLAG_DF;
			break;
		
		//Prefixes
		case 0x66:	//Operand Size Override
			DEBUG_S("PREFIX: OPRADND OVERRIDE");
			break;
		case 0x67:	//Memory Size Override
			DEBUG_S("PREFIX: ADDR OVERRIDE");
			break;
		
		default:
			//Short Jump
			if( (code[r->ip-1] & 0xF0) == 0x70)
			{
				DEBUG_S("J");
				pt2 = code[r->ip];
				switch(code[r->ip-1] & 0xF) {
				
				case 0:	//Overflow
					DEBUG_C('O');
					if(r->flags & FLAG_OF)	r->ip += code[r->ip];
					break;
				case 1:	//No Overflow
					DEBUG_S("NO");
					if(r->flags ^ FLAG_OF)	r->ip += code[r->ip];
					break;
				case 2:	//Carry
					DEBUG_C('C');
					if(r->flags & FLAG_CF)	r->ip += code[r->ip];
					break;
				case 3:	//No Carry
					DEBUG_S("NC");
					if(r->flags ^ FLAG_CF)	r->ip += code[r->ip];
					break;	
				case 4:	//Equal
					DEBUG_S("Z");
					if(r->flags & FLAG_ZF)	r->ip += code[r->ip];
					break;
				case 5:	//Not Equal
					DEBUG_S("NZ");
					if(!(r->flags & FLAG_ZF))	r->ip += code[r->ip];
					break;
				case 6:	//Below or Equal
					DEBUG_S("BE");
					if(r->flags & FLAG_CF || r->flags & FLAG_ZF)	r->ip += code[r->ip];
					break;
				case 7:	//Above
					DEBUG_S("A");
					if(r->flags ^ FLAG_CF && r->flags ^ FLAG_ZF)	r->ip += code[r->ip];
					break;
				
				}
				DEBUG_S(" (S) .+0x");
				DEBUG_H(pt2);
				
				r->ip++;
				break;
			}
			puts("Unkown Opcode IP = 0x");	putHex(code[r->ip-1]);
			__asm__ __volatile__ ("sti");
			for(;;)
				__asm__ __volatile__ ("hlt");
			return;
			break;
		}
		DEBUG_C('\n');
		#if DEBUG && defined(DELAY)
		#if DELAY == -1
		while(!kb_press);
		while(kb_press);
		#else
		for(pt2=DELAY;pt2--;);
		#endif
		#endif
	}
}

Uint8 *regB(t_regsRM *r, int code)
{
	switch(code) {
	case 0:
		DEBUG_S(" AL");
		return (Uint8*)&r->ax;
	case 1:
		DEBUG_S(" CL");
		return (Uint8*)&r->cx;
	case 2:
		DEBUG_S(" DL");
		return (Uint8*)&r->dx;
	case 3:
		DEBUG_S(" BL");
		return (Uint8*)&r->bx;
	case 4:
		DEBUG_S(" AH");
		return (Uint8*)(&r->ax)+1;
	case 5:
		DEBUG_S(" CH");
		return (Uint8*)(&r->cx)+1;
	case 6:
		DEBUG_S(" DH");
		return (Uint8*)(&r->dx)+1;
	case 7:
		DEBUG_S(" BH");
		return (Uint8*)(&r->bx)+1;
	}
	return 0;
}

Uint16 *regW(t_regsRM *r, int code)
{
	switch(code) {
	case 0:
		DEBUG_S(" AX");
		return &r->ax;
	case 1:
		DEBUG_S(" CX");
		return &r->cx;
	case 2:
		DEBUG_S(" DX");
		return &r->dx;
	case 3:
		DEBUG_S(" BX");
		return &r->bx;
	case 4:
		DEBUG_S(" SP");
		return &r->sp;
	case 5:
		DEBUG_S(" BP");
		return &r->bp;
	case 6:
		DEBUG_S(" SI");
		return &r->si;
	case 7:
		DEBUG_S(" DI");
		return &r->di;
	}
	return 0;
}

Uint16 *seg(t_regsRM *r, int code)
{
	switch(code) {
	case 0:
		DEBUG_S(" ES");
		return &r->es;
	case 1:
		DEBUG_S(" CS");
		return &r->cs;
	case 2:
		DEBUG_S(" SS");
		return &r->ss;
	case 3:
		DEBUG_S(" DS");
		return &r->ds;
	}
	return 0;
}

Uint16 *doFunc(t_regsRM *r, int mmm, int disp)
{
	switch(mmm) {
	case 0:
		DEBUG_F(" DS:[BX+SI+0x%x]", disp);
		return (Uint16 *)( (r->ds<<4) + r->bx + r->si + disp );
	case 1:
		DEBUG_F(" DS:[BX+DI+0x%x]", disp);
		return (Uint16 *)( (r->ds<<4) + r->bx + r->di + disp );
	case 2:
		DEBUG_F(" SS:[BP+SI+0x%x]", disp);
		return (Uint16 *)( (r->ss<<4) + r->bp + r->si + disp );
	case 3:
		DEBUG_F(" SS:[BP+DI+0x%x]", disp);
		return (Uint16 *)( (r->ss<<4) + r->bp + r->di + disp );
	case 4:
		DEBUG_F(" DS:[SI+0x%x]", disp);
		return (Uint16 *)( (r->ds<<4) + r->si + disp );
	case 5:
		DEBUG_F(" DS:[DI+0x%x]", disp);
		return (Uint16 *)( (r->ds<<4) + r->di + disp );
	case 6:
		DEBUG_F(" SS:[BP+0x%x]", disp);
		return (Uint16 *)( (r->ss<<4) + r->bp + disp );
	case 7:
		DEBUG_F(" DS:[BX+0x%x]", disp);
		return (Uint16 *)( (r->ds<<4) + r->bx + disp );
	}
	return 0;
}

int genToFromB(t_regsRM *r, Uint8 *data, Uint8 **to, Uint8 **from)
{
	int	d = *data;
	switch((d >> 6)&3) {
	case 0:	//No Offset
		if(to) *to = regB( r, (d>>3) & 7 );
		if(from) *from = (Uint8 *) doFunc( r, d & 7, 0 );
		return 1;
	case 1:	//8 Bit
		if(to) *to = regB( r, (d>>3) & 7 );
		if(from) *from = (Uint8 *) doFunc( r, d & 7, (data[1]&0x80?0xFF00:0)|data[1] );
		return 2;
	case 2:	//16 Bit
		if(to) *to = regB( r, (d>>3) & 7 );
		if(from) *from = (Uint8 *) doFunc( r, d & 7, *(Uint16*)(data+2) );
		return 3;
	case 3:	//Regs Only
		if(to) *to = regB( r, (d>>3) & 7 );
		if(from) *from = regB( r, d & 7 );
		return 1;
	}
	return 0;
}

int genToFromW(t_regsRM *r, Uint8 *data, Uint16 **to, Uint16 **from)
{
	int	d = *data;
	switch(d >> 6) {
	case 0:	//No Offset
		if(to) *to = regW( r, (d>>3) & 7 );
		if(from) *from = doFunc( r, d & 7, 0 );
		return 1;
	case 1:	//8 Bit
		if(to) *to = regW( r, (d>>3) & 7 );
		if(from) *from = doFunc( r, d & 7, (data[1]&0x80?0xFF00:0)|data[1] );
		return 2;
	case 2:	//16 Bit
		if(to) *to = regW( r, (d>>3) & 7 );
		if(from) *from = doFunc( r, d & 7, *(Uint16*)(data+2) );
		return 3;
	case 3:	//Regs Only
		if(to) *to = regW( r, (d>>3) & 7 );
		if(from) *from = regW( r, d & 7 );
		return 1;
	}
	return 0;
}

void doCJMP(t_regsRM *r, Uint8 type, Uint16 offset)
{
	DEBUG_S("J");
	switch(type) {
	case 0:	DEBUG_C('O');	//Overflow
		if(r->flags & FLAG_OF)	r->ip += offset;
		break;
	case 1:	DEBUG_S("NO");	//No Overflow
		if(r->flags ^ FLAG_OF)	r->ip += offset;
		break;
	case 2:	DEBUG_C('C');	//Carry
		if(r->flags & FLAG_CF)	r->ip += offset;
		break;
	case 3:	DEBUG_S("NC");	//No Carry
		if(r->flags ^ FLAG_CF)	r->ip += offset;
		break;	
	case 4:	DEBUG_S("Z");	//Equal
		if(r->flags & FLAG_ZF)	r->ip += offset;
		break;
	case 5:	DEBUG_S("NZ");	//Not Equal
		if(r->flags ^ FLAG_ZF)	r->ip += offset;
		break;
	case 6:	DEBUG_S("BE");	//Below or Equal
		if(r->flags & FLAG_CF || r->flags & FLAG_ZF)	r->ip += offset;
		break;
	case 7:	DEBUG_S("A");	//Above
		if(r->flags ^ FLAG_CF && r->flags ^ FLAG_ZF)	r->ip += offset;
		break;
	
	}
	DEBUG_S(" NEAR .+0x");	DEBUG_H(offset);
}
